
%{
    #include <stdio.h>
    #include "basic.h"
    
    extern StatementList *top;

%}

%union {
/*    struct ast *a;*/
    char *identifier;
    Expression *expression;
    ExpressionList *expression_list;
    Statement *statement;
    StatementList *statement_list;
    
/*    struct symbol *s;*/
/*    struct symlist *sl;*/
}

/* declare tokens */

%token ADD SUB MUL DIV MOD
%token COLOR
%token EQUAL
%token COLON SEMICOLON
%token LRB RRB
%token GT LT
%token OR NOT AND XOR
%token NE GE LE
%token DIM
%token END READ DATA RESTORE
%token IF THEN
%token ON
%token FOR TO STEP NEXT
%token GOSUB RETURN
%token GOTO
%token VSYNC

%token ASC CHR VAL STR HEX MID LEN

%token PRINT

%token RND ABS SGN PI RAD DEG SIN COS TAN ATAN SQR EXP LOG FLOOR

%token INPUT

%token NEWLINE COMMA

%token <identifier> IDENTIFIER
%token <identifier> IDENTIFIER_STRING
%token <identifier> LABEL

%token <expression> INT_LITERAL DOUBLE_LITERAL STRING_LITERIAL

%type <expression> constant_literial
%type <expression> expression primary_identifier
%type <expression> primary_array_identifier
%type <expression> primary_print_identifier

%type <expression_list> constant_literial_list primary_identifier_list primary_array_identifier_list
%type <expression_list> print_expression_list print_list
%type <expression_list> label_list


%type <statement_list> program lines statements statements2
%type <statement> statement statement2 if_statement for_statement gosub_statement
%type <statement> goto_statement

%right EQUAL
%left ADD SUB
%left MUL DIV MOD
%left OR AND XOR
%left EQ NE GT LT GE LE
%left IF THEN

%nonassoc UMINUS
%nonassoc NOT

%start program

%%

program:
 statements
{
        top = $1;
}
| lines
{
}
| lines statements
{
    chain_statements($1, $2);

}
;

lines:
 statements NEWLINE
{
        top = $1;
}
| lines statements NEWLINE
{
    $$ = chain_statements($1, $2);

/*
    printf("lines: %x - %x : ", $1->statement, $2->statement);

    if ($1->statement->type == PRINT_STATEMENT) {
        printf("print: %x (%d) ", $1->statement, $2->statement->u.expression_s->u.int_value);
    }
    if ($2->statement->type == PRINT_STATEMENT) {
        printf("print: %x (%d)\n", $2, $2->statement->u.expression_s->u.int_value);
    }

    if ($2->statement->type == IF_STATEMENT) {
        printf("lines chain IF: %x\n", $2);
    }
*/
}
;

statements:
 statement
{
    $$ = create_statement_list($1);
/*
    printf("chain print:%x\n", $1);
*/
}
| statements COLON statement
{
    $$ = chain_statement_list($1, $3);

/*
    printf("chain: %x - %x : ", $1->statement, $3);

    if ($1->statement->type == PRINT_STATEMENT) {
        printf("print: %x (%d) ", $1->statement, $1->statement->u.expression_s->u.int_value);
    }
    if ($3->type == PRINT_STATEMENT) {
        printf("print: %x (%d)\n", $3, $3->u.expression_s->u.int_value);
    }

    if ($3->type == IF_STATEMENT) {
        printf("chain IF: %x\n", $3);
    }
*/
}
;

statement:
 statement2
{
    $$ = $1;
}
| LABEL
{
    Statement *a = create_alloc_statement();
    a->type = LABEL_STATEMENT;
    a->u.label_s = $1;
    $$=a;
}
;

statements2:
 statement2
{
    $$ = create_statement_list($1);
}
| statements2 COLON statement2
{
    $$ = chain_statement_list($1, $3);
}
;

statement2:
{
    $$ = create_empty_statement();
}
| PRINT print_list
{ 
    $$ = create_print_statement($2);
}
| primary_identifier EQUAL expression
{
/*        printf("%s = \n", $1);*/

    $$ = create_assign_statement($1, $3);
}
| DIM primary_array_identifier_list
{
    $$ = create_dim_statement($2);
}
| if_statement
{
    $$ = $1;
}
| for_statement
{
    $$ = $1;
}
| NEXT primary_identifier
{
    $$ = create_next_var_statement($2);
}
| NEXT 
{
    $$ = create_next_statement();
}
| END
{
    Statement *a = create_alloc_statement();
    a->type = END_STATEMENT;
    $$=a;
}
| VSYNC expression
{
    $$ = create_vsync_statement($2);
}
| COLOR expression
{
    $$ = create_color_statement($2);
}
| gosub_statement
{
    $$ = $1;
}
| goto_statement
{
    $$ = $1;
}
| RETURN
{
    $$ = create_return_statement();
}
| READ primary_identifier_list
{
    $$ = create_read_statement($2);
}
| DATA constant_literial_list
{
    $$ = create_data_statement($2);
}
| RESTORE LABEL
{
    $$ = create_restore_statement($2); 
}
| INPUT primary_identifier_list
{
    $$ = create_input_statement(NULL, $2);
}
| INPUT primary_print_identifier SEMICOLON primary_identifier_list
{
    ExpressionList *a;

    Expression  *expression = create_alloc_expression();
    expression->type = SEMICOLON_EXPRESSION;
    expression->u.string_value = create_identifier(";");

    a = create_array_expression($2);
    a = chain_array_expression(a, expression); 

    $$ = create_input_statement(a, $4);
}
| ON expression GOTO label_list
{
    $$ = create_on_goto_statement($2, $4);
}
| ON expression GOSUB label_list
{
    $$ = create_on_gosub_statement($2, $4);
}
;

print_expression_list:
 expression
{
    $$ = create_array_expression($1);
}
| print_expression_list COMMA expression
{
    ExpressionList *a;
    Expression  *expression = create_alloc_expression();
    expression->type = COMMA_EXPRESSION;
    expression->u.string_value = create_identifier(",");

    a = chain_array_expression($1, expression); 
    $$ = chain_array_expression(a, $3);    
}
| print_expression_list SEMICOLON expression
{
    ExpressionList *a;
    Expression  *expression = create_alloc_expression();
    expression->type = SEMICOLON_EXPRESSION;
    expression->u.string_value = create_identifier(";");

    a = chain_array_expression($1, expression); 
    $$ = chain_array_expression(a, $3);    
}
;

print_list:
 print_expression_list COMMA
{
    Expression  *expression = create_alloc_expression();
    expression->type = COMMA_EXPRESSION;
    expression->u.string_value = create_identifier(",");

    $$ = chain_array_expression($1, expression); 
}
| print_expression_list SEMICOLON
{
    Expression  *expression = create_alloc_expression();
    expression->type = SEMICOLON_EXPRESSION;
    expression->u.string_value = create_identifier(",");

    $$ = chain_array_expression($1, expression); 
}
| print_expression_list
{
    Expression  *expression = create_alloc_expression();
    expression->type = EMPTY_EXPRESSION;
    expression->u.string_value = create_identifier(" ");

    $$ = chain_array_expression($1, expression); 
}
;

if_statement:
 IF expression THEN statements2
{
/*
        printf("type:%d\n", $4->statement->type);
        printf("type:%x\n", $4->next);
*/
        $$ = create_if_statement($2, $4);
/*
        printf("type:%d\n", $$->type);
*/
}
| IF expression THEN LABEL
{
    Statement *statement;
    StatementList *statement_list;

    statement = create_goto_statement($4);
    statement_list = create_statement_list(statement);

    $$ = create_if_statement($2, statement_list);  
}
| IF expression GOTO LABEL
{
    Statement *statement;
    StatementList *statement_list;
    
    statement = create_goto_statement($4);
    statement_list = create_statement_list(statement);
    
    $$ = create_if_statement($2, statement_list);  
}
;

for_statement:
 FOR primary_identifier EQUAL expression TO expression STEP expression
{
    $$ = create_for_statement($2, $4, $6, $8);
}
|  FOR primary_identifier EQUAL expression TO expression
{    
    Expression  *expression = create_alloc_expression();
    expression->type = INT_EXPRESSION;
    expression->u.int_value = 1;

    $$ = create_for_statement($2, $4, $6, expression);
}
;

gosub_statement:
 GOSUB LABEL
{
    $$ = create_gosub_statement($2);
}
;

goto_statement:
 GOTO LABEL
{
    $$ = create_goto_statement($2);
}
;

expression: 
  expression ADD expression     { $$ = create_binary_expression(ADD_EXPRESSION, $1, $3); }
| expression SUB expression     { $$ = create_binary_expression(SUB_EXPRESSION, $1, $3); }
| expression MUL expression     { $$ = create_binary_expression(MUL_EXPRESSION, $1, $3); }
| expression DIV expression     { $$ = create_binary_expression(DIV_EXPRESSION, $1, $3); }
| expression MOD expression     { $$ = create_binary_expression(MOD_EXPRESSION, $1, $3); }
| SUB expression %prec UMINUS 
{ 
    $$ = create_minus_expression($2);
}
| expression OR expression
{
    $$ = create_binary_expression(OR_EXPRESSION, $1, $3);
}
|   NOT expression
{
    /*
     if ($2->type == INT_EXPRESSION) {
     printf("NOT int(%d)\n", $2->u.int_value);
     }
     */    
    $$ = create_not_expression($2);
}
|   expression AND expression
{ 
    $$ = create_binary_expression(AND_EXPRESSION, $1, $3);
}
|   expression XOR expression
{ 
    $$ = create_binary_expression(XOR_EXPRESSION, $1, $3);
}
|   expression EQ expression
{ 
    $$ = create_binary_expression(EQ_EXPRESSION, $1, $3);
}
|   expression NE expression
{
    $$ = create_binary_expression(NE_EXPRESSION, $1, $3);
}
|   expression GT expression
{
    $$ = create_binary_expression(GT_EXPRESSION, $1, $3); 
}
|   expression LT expression
{ 
    $$ = create_binary_expression(LT_EXPRESSION, $1, $3); 
}
|   expression GE expression
{ 
    $$ = create_binary_expression(GE_EXPRESSION, $1, $3); 
}
|   expression LE expression
{ 
    $$ = create_binary_expression(LE_EXPRESSION, $1, $3); 
}
|   LRB expression RRB
{
    $$ = $2;
}
| ASC LRB expression RRB
{
    $$ = create_asc_expression($3);     
}
| CHR LRB expression RRB
{
    $$ = create_chr_expression($3);     
}
| VAL LRB expression RRB
{
    $$ = create_val_expression($3);     
}
| STR LRB expression RRB
{
    $$ = create_str_expression($3);     
}
| MID LRB expression COMMA expression COMMA expression RRB
{
    $$ = create_mid_expression($3, $5, $7);     
}
| HEX LRB expression RRB
{
    $$ = create_hex_expression($3);     
}
| LEN LRB expression RRB
{
    $$ = create_len_expression($3);     
}
| RND LRB expression RRB
{
    $$ = create_rnd_expression($3);     
}
| ABS LRB expression RRB
{
    $$ = create_abs_expression($3);
}
| SGN LRB expression RRB
{
    $$ = create_sgn_expression($3);
}
| PI LRB RRB
{
    $$ = create_pi_expression();
}
| RAD LRB expression RRB
{
    $$ = create_rad_expression($3);
}
| DEG LRB expression RRB
{
    $$ = create_deg_expression($3);
}
| SIN LRB expression RRB
{
    $$ = create_sin_expression($3);
}
| COS LRB expression RRB
{
    $$ = create_cos_expression($3);
}
| TAN LRB expression RRB
{
    $$ = create_tan_expression($3);
}
| ATAN LRB expression RRB
{
    $$ = create_atan_expression($3);
}
| ATAN LRB expression COMMA expression RRB
{
    $$ = create_atan_yx_expression($3, $5);
}
| SQR LRB expression RRB
{
    $$ = create_sqr_expression($3);
}
| EXP LRB expression RRB
{
    $$ = create_exp_expression($3);
}
| LOG LRB expression RRB
{
    $$ = create_log_expression($3);
}
| FLOOR LRB expression RRB
{
    $$ = create_floor_expression($3);
}
| INT_LITERAL
{
    $$ = $1;
}
| DOUBLE_LITERAL
{
    $$ = $1;
}
| STRING_LITERIAL
{
    $$ = $1;
}
| primary_identifier
;


primary_identifier:
 IDENTIFIER             
{
    $$ = create_identifier_expression($1, IDENTIFIER_VALUE_INT, IDENTIFIER_DIMENSION0, NULL, NULL);
}
| IDENTIFIER_STRING
{
    $$ = create_identifier_expression($1, IDENTIFIER_VALUE_STRING, IDENTIFIER_DIMENSION0, NULL, NULL);            
}
| primary_array_identifier
;



primary_array_identifier:
 IDENTIFIER LRB expression RRB
{
    $$ = create_identifier_expression($1, IDENTIFIER_VALUE_NUMBER_ARRAY, IDENTIFIER_DIMENSION1, $3, NULL);    
}
| IDENTIFIER LRB expression COMMA expression RRB
{
    $$ = create_identifier_expression($1, IDENTIFIER_VALUE_NUMBER_ARRAY, IDENTIFIER_DIMENSION2, $3, $5);        
}
| IDENTIFIER_STRING LRB expression RRB
{
    $$ = create_identifier_expression($1, IDENTIFIER_VALUE_STRING_ARRAY, IDENTIFIER_DIMENSION1, $3, NULL);                
}
| IDENTIFIER_STRING LRB expression COMMA expression RRB
{
    $$ = create_identifier_expression($1, IDENTIFIER_VALUE_STRING_ARRAY, IDENTIFIER_DIMENSION2, $3, $5);            
}
;

primary_identifier_list:
 primary_identifier
{
    $$ = create_array_expression($1);
    
}
| primary_identifier_list COMMA primary_identifier
{
    $$ = chain_array_expression($1, $3);    
}
;


primary_array_identifier_list:
 primary_array_identifier
{
    $$ = create_array_expression($1);

}
| primary_array_identifier_list COMMA primary_array_identifier
{
    $$ = chain_array_expression($1, $3);    
}
;

primary_print_identifier:
 STRING_LITERIAL
{
    $$ = $1;
}
| IDENTIFIER_STRING
{
    $$ = create_identifier_expression($1, IDENTIFIER_VALUE_STRING, IDENTIFIER_DIMENSION0, NULL, NULL);            
}
| IDENTIFIER_STRING LRB expression RRB
{
    $$ = create_identifier_expression($1, IDENTIFIER_VALUE_STRING_ARRAY, IDENTIFIER_DIMENSION1, $3, NULL);                
}
| IDENTIFIER_STRING LRB expression COMMA expression RRB
{
    $$ = create_identifier_expression($1, IDENTIFIER_VALUE_STRING_ARRAY, IDENTIFIER_DIMENSION2, $3, $5);            
}
;

label_list:
 LABEL
{
    Expression *expression;
    
    expression = create_lable_expression($1);
    $$ = create_array_expression(expression);
}
| label_list COMMA LABEL
{
    Expression *expression;

    expression = create_lable_expression($3);
    $$ = chain_array_expression($1, expression);
}
;

constant_literial:
 INT_LITERAL
{
    $$ = $1;
}
| DOUBLE_LITERAL
{
    $$ = $1;
}
| SUB INT_LITERAL
{
    $2->u.int_value = -1 * $2->u.int_value;
    $$ = $2;
}
| SUB DOUBLE_LITERAL
{
    if ($2->u.double_value != 0.0) {
        $2->u.double_value = -1 * $2->u.double_value;
    }
    $$ = $2;
}
| STRING_LITERIAL
{
    $$ = $1;
}
;

constant_literial_list:
 constant_literial
{
    $$ = create_array_expression($1);
}
| constant_literial_list COMMA constant_literial
{
    $$ = chain_array_expression($1, $3);   
}
;

%%
