#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include "basic.h"
#include "error.h"

#pragma mark -
#pragma mark Expression

/*
    Expression
 */

Expression *create_alloc_expression()
{
    Expression *a = malloc(sizeof(Expression));

    a->line_number = yylineno;
    
    if(!a) {
        yyerror(runtime_error_message[OUT_OF_MEMORY_SPACE_ERR].format);
    }

    return a;
}

Expression *create_minus_expression(Expression *exp) {
    Expression *a =  create_alloc_expression();
    
    a->type = MINUS_EXPRESSION;
    a->u.expression = exp;

    return a;
}


Expression *create_binary_expression(ExpressionType type, Expression *left, Expression *right)
{
    Expression *a =  create_alloc_expression();
    a->type = type;
    a->u.binary_expression.left = left;
    a->u.binary_expression.right = right;

    return a;
}


Expression *create_not_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();

    a->type = NOT_EXPRESSION;
    a->u.bit_not = exp;
    
    return a;
    
}


Expression *create_identifier_expression(char *name, IdentifierValueType valueType, IdentifierDimensionType dimensionType, Expression *dimemsion1, Expression *dimension2) {

    Expression *a =  create_alloc_expression();
    
    a->type = IDENTIFIER_EXPRESSION;
    a->u.identifier.identifier = name;
    a->u.identifier.valueType = valueType;
    a->u.identifier.dimensionType = dimensionType;
    a->u.identifier.dimension1 = dimemsion1;
    a->u.identifier.dimension2 = dimension2;

    
    return a;
}


ExpressionList *create_alloc_expression_List() {
    ExpressionList *a = malloc(sizeof(ExpressionList));
    
    if(!a) {
        yyerror(runtime_error_message[OUT_OF_MEMORY_SPACE_ERR].format);
    }
    
    return a; 

}
 
ExpressionList *create_array_expression(Expression *exp) {
    ExpressionList *a = create_alloc_expression_List();
    a->expression = exp;
    a->next = NULL;
    
    /*    StatementList *next = tail_statement_list(); 
     next = a;
     */
    return a;
 
}

ExpressionList *chain_array_expression(ExpressionList *list, Expression *exp) {
    ExpressionList *a;
    ExpressionList *p;

    a = malloc(sizeof(ExpressionList));
    
    a->expression = exp;
    a->next = NULL;

    p = list;
    
    while (p->next != NULL) {
        p = (ExpressionList *)p->next;
    }
        
    p->next = (struct ExpressionList_tag *)a;
    
    return list;
  
}

Expression *create_asc_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = ASC_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_chr_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = CHR_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_val_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = VAL_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_str_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = STR_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_mid_expression(Expression *source, Expression *location, Expression *size)
{
    Expression *a =  create_alloc_expression();
    
    a->type = MID_EXPRESSION;
    a->u.mid_expression.source = source;
    a->u.mid_expression.location = location;
    a->u.mid_expression.size = size;
    
    return a;
    
}

Expression *create_hex_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = HEX_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_len_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = LEN_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_rnd_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = RND_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_lable_expression(char *label)
{
    Expression *a =  create_alloc_expression();
    
    a->type = LABEL_EXPRESSION;
    a->u.label = label;
    
    return a;
    
}

Expression *create_abs_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = ABS_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_sgn_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = SGN_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_pi_expression()
{
    Expression *a =  create_alloc_expression();
    
    a->type = DOUBLE_EXPRESSION;
    a->u.double_value = M_PI;
    
    return a;
    
}

Expression *create_rad_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = RAD_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_deg_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = DEG_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_sin_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = SIN_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_cos_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = COS_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_tan_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = TAN_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_atan_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = ATAN_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_atan_yx_expression(Expression *left, Expression *right)
{
    Expression *a =  create_alloc_expression();
    
    a->type = ATAN_YX_EXPRESSION;
    a->u.binary_expression.left = left;
    a->u.binary_expression.right = right;

    return a;
    
}

Expression *create_sqr_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = SQR_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_exp_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = EXP_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_log_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = LOG_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}

Expression *create_floor_expression(Expression *exp)
{
    Expression *a =  create_alloc_expression();
    
    a->type = FLOOR_EXPRESSION;
    a->u.expression = exp;
    
    return a;
    
}


#pragma mark -
#pragma mark Statement

/*
    Statement
 */ 

Statement *create_alloc_statement()
{
    Statement *a = malloc(sizeof(Statement));

    a->line_number = yylineno;
    a->flag = B_FALSE;

    if(!a) {
        yyerror(runtime_error_message[OUT_OF_MEMORY_SPACE_ERR].format);
    }
    
    return a;  
}
                                
Statement *create_expression_statement(Expression *expression)
{
    Statement *a = create_alloc_statement();
    a->u.expression_s = expression;
    
    return a;
}

StatementList *create_alloc_statement_List()
{
    StatementList *a = malloc(sizeof(StatementList));
    
    if(!a) {
        yyerror(runtime_error_message[OUT_OF_MEMORY_SPACE_ERR].format);
        exit(0);
    }
    
    return a; 
}

Statement *create_assign_statement(Expression *identifier, Expression *expression) {
    Statement *a = create_alloc_statement();
    
    a->type = ASSIGN_EXPRESSION_STATEMENT;
    a->u.assign_s.identifier = identifier; 
    a->u.assign_s.expression = expression;
    
    return a;    
}

Statement *create_print_statement(ExpressionList *array) {
    Statement *a = create_alloc_statement();
    
    a->type = PRINT_STATEMENT;
    a->u.array_s = array;
    
    return a;
    
}

Statement *create_dim_statement(ExpressionList *array) {
    Statement *a = create_alloc_statement();
    
    a->type = DIM_STATEMENT;
    a->u.array_s = array;
    
    return a;
  
}


Statement *create_if_statement(Expression *expression, StatementList *list)
{
    Statement *a = create_alloc_statement();

    a->type = IF_STATEMENT;
    a->u.if_s.condition = expression; 
    a->u.if_s.statement_list = list;    
    
    return a;
    
}

Statement *create_for_statement(Expression *identifier, Expression *expression_init, Expression *expression_to, Expression *expression_step) {
    Statement *a = create_alloc_statement();
    
    a->type = FOR_STATEMENT;
    a->u.for_s.identifier = identifier; 
    a->u.for_s.expression_init = expression_init;
    a->u.for_s.expression_to = expression_to;
    a->u.for_s.expression_step = expression_step;
    
    return a;    
}

Statement *create_next_statement() {
    Statement *a = create_alloc_statement();
    
    a->type = NEXT_STATEMENT; 
    
    return a;    
}

Statement *create_next_var_statement(Expression *var) {
    Statement *a = create_alloc_statement();
    
    a->type = NEXT_VAR_STATEMENT; 
    a->u.expression_s = var;
    
    return a;    
}


Statement *create_gosub_statement(char *label) {
    Statement *a = create_alloc_statement();
    
    a->type = GOSUB_STATEMENT; 
    a->u.label_s = label;
    
    return a;    
}

Statement *create_goto_statement(char *label) {
    Statement *a = create_alloc_statement();
    
    a->type = GOTO_STATEMENT; 
    a->u.label_s = label;
    
    return a;    
}

Statement *create_restore_statement(char *label) {
    Statement *a = create_alloc_statement();
    
    a->type = RESTORE_STATEMENT; 
    a->u.label_s = label;
    
    return a;    
}


Statement *create_return_statement() {
    Statement *a = create_alloc_statement();
    
    a->type = RETURN_STATEMENT; 
    
    return a;    
}

Statement *create_empty_statement() {
    Statement *a = create_alloc_statement();
    
    a->type = EMPTY_STATEMENT; 
    
    return a;    
}


StatementList *create_statement_list(Statement *statement)
{
    StatementList *a = create_alloc_statement_List();
    a->statement = statement;
    a->next = NULL;
    
    /*    StatementList *next = tail_statement_list(); 
     next = a;
     */
    return a;
}

StatementList *chain_statement_list(StatementList *list, Statement *statement)
{   
    StatementList *a = create_statement_list(statement);
    StatementList *p;

    p = list;

    while (p->next != NULL) {
        p = (StatementList *)p->next;
    }
    
   p->next = (struct StementList_tag *)a;
    
    return list;
}


StatementList *chain_statements(StatementList *left, StatementList *right)
{   
    StatementList *p;
    
    p = left;
    
    while (p->next != NULL) {
        p = (StatementList *)p->next;
    }
    
    p->next = (struct StementList_tag *)right;
    
    return left;
}

Statement *create_read_statement(ExpressionList *array) {
    Statement *a = create_alloc_statement();
    
    a->type = READ_STATEMENT;
    a->u.array_s = array;
    
    return a;
    
}

Statement *create_data_statement(ExpressionList *array) {
    Statement *a = create_alloc_statement();
    
    a->type = DATA_STATEMENT;
    a->u.array_s = array;
    
    return a;
    
}

Statement *create_input_statement(ExpressionList *print_array, ExpressionList *array) {    
    Statement *a = create_alloc_statement();
    
    a->type = INPUT_STATEMENT;
    a->u.input_s.print_array_s = print_array;
    a->u.input_s.array_s = array;
    
    return a;    
}

Statement *create_on_goto_statement(Expression *expression, ExpressionList *array) {
    Statement *a = create_alloc_statement();
    
    a->type = ON_GOTO_STATEMENT; 
    a->u.on_goto_s.expression_s = expression;
    a->u.on_goto_s.array_s = array;
        
    return a;    
}

Statement *create_on_gosub_statement(Expression *expression, ExpressionList *array) {
    Statement *a = create_alloc_statement();
    
    a->type = ON_GOSUB_STATEMENT; 
    a->u.on_gosub_s.expression_s = expression;
    a->u.on_gosub_s.array_s = array;
    
    return a;    
}

Statement *create_color_statement(Expression *expression) {
    Statement *a = create_alloc_statement();
    
    a->type = COLOR_STATEMENT; 
    a->u.expression_s = expression;
    
    return a;    
}

Statement *create_vsync_statement(Expression *expression) {
    Statement *a = create_alloc_statement();
    
    a->type = VSYNC_STATEMENT; 
    a->u.expression_s = expression;
    
    return a;    
}


