#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "basic.h"
#include "error.h"

/*
    Symbol
 */

EvalValueList *create_alloc_eval_value_list() {
    EvalValueList *a = malloc(sizeof(EvalValueList));
    
    if(!a) {
        runtime_error_no_line(OUT_OF_MEMORY_SPACE_ERR);
        exit(0);
    }
    
    return a; 
}

EvalValueList *create_init_eval_value_list(IdentifierValueType type, int index) {
    EvalValueList *a;
    
    a = create_alloc_eval_value_list();
    
    a->value.type = type;
    a->value.index = index;
    
    switch (a->value.type) {
        case IDENTIFIER_VALUE_INT:
            a->value.u.int_value = 0;
            break;
        case IDENTIFIER_VALUE_DOUBLE:
            a->value.u.double_value = 0;
            break;
        case IDENTIFIER_VALUE_STRING:
            a->value.u.string_value = sym_string_empty;
            break;
    }

    a->next = NULL;

    return a;
}

void chain_eval_value_list(Symbol *sym, EvalValueList *list) {    
    EvalValueList *a;
    
    a = sym->u.array;
    
    
    if (a == NULL) {
        sym->u.array = list;
    } else {
        while (a->next != NULL) {
            a = (struct EvalValueList_tag *)a->next;
        }
        a->next = list;
    }
    
}

SymList *create_alloc_init_symbol(Expression *identifier) {
    int i;
    EvalValueList *ptr;
    SymList *p;
    SymList *a;
    EvalValue eval_limit_dimension1;
    EvalValue eval_limit_dimension2;
    
    p = lookup(identifier->u.identifier.identifier, identifier->u.identifier.valueType);

    if (p != NULL) {
        runtime_error(identifier->line_number, DUPLICATE_DEFINE_DIM_VARIABLE_ERR);
        exit(1);
        return NULL;
    }
    
    a = malloc(sizeof(SymList));
    
    if(!a) {
        runtime_error(identifier->line_number, OUT_OF_MEMORY_SPACE_ERR);
        exit(0);
    }

    a->sym.name = strdup(identifier->u.identifier.identifier);

    a->sym.valueType = identifier->u.identifier.valueType;
    a->sym.dimensionType = identifier->u.identifier.dimensionType;

    switch (identifier->u.identifier.dimensionType) {
        case IDENTIFIER_DIMENSION1:
            eval_limit_dimension1 = eval_expression(identifier->u.identifier.dimension1);            
            
            if (eval_limit_dimension1.type == IDENTIFIER_VALUE_STRING) {
                runtime_error(identifier->line_number, VARIABLE_NOT_ASSIGN_ERR);
                exit(1);
            }
            
            if (eval_limit_dimension1.type == IDENTIFIER_VALUE_DOUBLE) {
                a->sym.limit_dimension1 = (int)eval_limit_dimension1.u.double_value;
            } else {
                a->sym.limit_dimension1 = eval_limit_dimension1.u.int_value;
            }            
            a->sym.limit_dimension2 = 0;
            
            if (a->sym.limit_dimension1 == 0) {                
                runtime_error(identifier->line_number, ARRAY_INDEX_OUT_OF_BOUNDS_ERR);
                exit(1);
            }
            break;
        case IDENTIFIER_DIMENSION2:
            eval_limit_dimension1 = eval_expression(identifier->u.identifier.dimension1);
            eval_limit_dimension2 = eval_expression(identifier->u.identifier.dimension2);
            
            if ((eval_limit_dimension1.type == IDENTIFIER_VALUE_STRING) || (eval_limit_dimension2.type == IDENTIFIER_VALUE_STRING)) {
                runtime_error(identifier->line_number, VARIABLE_NOT_ASSIGN_ERR);
                exit(1);
            }
            
            if (eval_limit_dimension1.type == IDENTIFIER_VALUE_DOUBLE) {
                a->sym.limit_dimension1 = (int)eval_limit_dimension1.u.double_value;
            } else {
                a->sym.limit_dimension1 = eval_limit_dimension1.u.int_value;
            }
            
            if (eval_limit_dimension2.type == IDENTIFIER_VALUE_DOUBLE) {
               a->sym.limit_dimension2 = (int)eval_limit_dimension2.u.double_value;
            } else {
               a->sym.limit_dimension2 = eval_limit_dimension2.u.int_value;
            }
            
            if (a->sym.limit_dimension1 == 0) {                
                runtime_error(identifier->line_number, ARRAY_INDEX_OUT_OF_BOUNDS_ERR);
                exit(1);
            }

            if (a->sym.limit_dimension2 == 0) {                
                runtime_error(identifier->line_number, ARRAY_INDEX_OUT_OF_BOUNDS_ERR);
                exit(1);
            }

            break;
    }

    Symbol *sym;
        
    sym = &(a->sym);

    switch (a->sym.valueType) {
        case IDENTIFIER_VALUE_INT:
            sym->u.value.type = IDENTIFIER_VALUE_INT;
            sym->u.value.index = 0;
            sym->u.value.u.int_value = 0;
            break;
        case IDENTIFIER_VALUE_DOUBLE:
            sym->u.value.type = IDENTIFIER_VALUE_DOUBLE;
            sym->u.value.index = 0;
            sym->u.value.u.double_value = 0.0;
            break;
        case IDENTIFIER_VALUE_STRING:
            sym->u.value.type = IDENTIFIER_VALUE_STRING;
            sym->u.value.index = 0;
            sym->u.value.u.string_value = sym_string_empty;
            break;
        case IDENTIFIER_VALUE_NUMBER_ARRAY:
            sym->u.array = NULL;
            switch (a->sym.dimensionType) {
                case IDENTIFIER_DIMENSION1:

                    for (i=0; i<a->sym.limit_dimension1; i++) {
                        ptr = create_init_eval_value_list(IDENTIFIER_VALUE_INT, i);
                         chain_eval_value_list(sym, ptr);
                    }
                    break;
                case IDENTIFIER_DIMENSION2:
                    for (i=0; i<(a->sym.limit_dimension1 * a->sym.limit_dimension2); i++) {
                        ptr = create_init_eval_value_list(IDENTIFIER_VALUE_INT, i);
                        chain_eval_value_list(sym, ptr);
                    }
                    break;
            }
            break;
        case IDENTIFIER_VALUE_STRING_ARRAY:
            switch (a->sym.dimensionType) {
                case IDENTIFIER_DIMENSION1:
                    for (i=0; i<a->sym.limit_dimension1; i++) {
                        ptr = create_init_eval_value_list(IDENTIFIER_VALUE_STRING, i);
                        chain_eval_value_list(sym, ptr);
                    }
                    break;
                case IDENTIFIER_DIMENSION2:
                    for (i=0; i<(a->sym.limit_dimension1 * a->sym.limit_dimension2); i++) {
                        ptr = create_init_eval_value_list(IDENTIFIER_VALUE_STRING, i);
                        chain_eval_value_list(sym, ptr);
                    }
                    break;
            }
    }
    
    return a;
   
}

SymList *lookup(char *name, IdentifierValueType type) {
    SymList *p;
    
    p=sym_list;

    while (p != NULL) {
        switch (type) {
            case IDENTIFIER_VALUE_INT:
            case IDENTIFIER_VALUE_DOUBLE:
                if (strcmp(p->sym.name, name) == 0) {
                    return p;
                }
                break;              
            case IDENTIFIER_VALUE_STRING:
            case IDENTIFIER_VALUE_NUMBER_ARRAY:
            case IDENTIFIER_VALUE_STRING_ARRAY:  
                if ((strcmp(p->sym.name, name) == 0) && (p->sym.valueType == type)) {
                    return p;
                }
                break;              
        }
        
        
        p = p->next;
    }

    return NULL;
}

void create_dim_symbol(ExpressionList *identifier_list) {
    Expression *identifier;
    ExpressionList *list;
    SymList *a;
    SymList *b;
    
    list = identifier_list;
    
    while (list != NULL) {
        a = create_alloc_init_symbol(list->expression);      

        if (a == NULL) {
            runtime_error(list->expression->line_number, DUPLICATE_DEFINE_DIM_VARIABLE_ERR);
            exit(1);
            return;
        }
        
        b = sym_list;
        
        if (b != NULL) {
            while (b->next != NULL) {
                b = b->next;
            }
            b->next = a;
        } else {
            sym_list = a;
        }

        list = (struct ExpressionList_tag *)list->next;
    }    
}

void create_symbol(Expression *identifier) {
    SymList *b;
    SymList *a;
    
    a = create_alloc_init_symbol(identifier);

    if (a == NULL) {
        return;
    }
    
    b = sym_list;
    
    if (b != NULL) {
        while (b->next != NULL) {
            b = b->next;
        }
        b->next = a;
    } else {
        sym_list = a;
    }
}


/*
 static unsigned
 symhash(char *sym)
 {
 unsigned int hash = 0;
 unsigned c;
 
 while(c = *sym++) hash = hash*9 ^ c;
 
 return hash;
 }
 */


/*
    Global
 */

StatementList *top = NULL;
StatementList *lastPointer = NULL;


ForStack forStacks[STACK_BUFFER_SIZE];
int for_stack_size = 0;


GosubStack gosubStacks[GOSUB_BUFFER_SIZE];
int gosub_stack_size = 0;


SymList *sym_list = NULL;
char *sym_string_empty = "";
int string_literial_buffer_size = 0; 
char *string_literial_buffer = NULL;

DataPointer dataPointer;

ExpressionType print_type;

StatementList *executePointer = NULL;
IFStack ifStacks[IF_BUFFER_SIZE];
int if_stack_size = 0;

Boolean if_stack = B_FALSE;
int if_stack_count = 0;

Boolean search_next = B_FALSE;
int search_next_count = 0;

/*
 
 */
void yyerror(char *s, ...)
{
    va_list ap;
    va_start(ap, s);
    
    print_esc();
    
    fprintf(stderr, "%d: ", yylineno);
    if (!strcmp(s,"syntax error")) {
        fprintf(stderr, "%s", runtime_error_message[SYNTAX_ERR].format);
    } else {
        vfprintf(stderr, s, ap);
    }
    fprintf(stderr, "\n");

/*
    StatementList *p;
    
    p=top;
    while (p!=NULL) {
        printf("%d\n", p->statement->type);
        p = p->next;
    }
*/    
    exit(1);
}


